import pytest

from cacholote import decode


def test_import_object() -> None:
    # Goedel-style self-reference :)
    res = decode.import_object("test_10_decode:test_import_object")
    assert res is test_import_object

    res = decode.import_object("builtins:len")
    assert res is len

    with pytest.raises(
        ValueError, match=r"'builtins.len' not in the form 'module:qualname'"
    ):
        decode.import_object("builtins.len")


def test_object_hook() -> None:
    object_simple = {
        "type": "python_object",
        "fully_qualified_name": "builtins:OSError",
    }
    res = decode.object_hook(object_simple)
    assert res is OSError

    len_simple = {"type": "python_object", "fully_qualified_name": "builtins:len"}
    res = decode.object_hook(len_simple)
    assert res is len

    len_call_simple = {"type": "python_call", "callable": len, "args": ["test"]}
    res = decode.object_hook(len_call_simple)
    assert res == len("test")

    len_call_simple = {
        "type": "python_call",
        "callable": "builtins:len",
        "args": ["test"],
    }
    res = decode.object_hook(len_call_simple)
    assert res == len("test")

    unsupported_object = {"key": 1}
    res = decode.object_hook(unsupported_object)
    assert res is unsupported_object

    unsupported_type = {"type": "unsupported_type"}
    res = decode.object_hook(unsupported_type)
    assert res is unsupported_type


def test_loads() -> None:
    len_call_json = (
        r'{"type":"python_call",'
        r'"callable":{"type":"python_object","fully_qualified_name":"builtins:int"}}'
    )

    res = decode.loads(len_call_json)
    assert res == int()



import contextlib
import os
import pathlib
import shlex
import subprocess
import time
from typing import Any, Dict, Generator

import psycopg
import pytest

from cacholote import config


def wait_s3_up(
    endpoint_url: str, max_sleep: float = 1.0, min_sleep: float = 0.1
) -> None:
    requests = pytest.importorskip("requests")
    while max_sleep > 0:
        time.sleep(min_sleep)
        max_sleep -= min_sleep
        try:
            r = requests.get(endpoint_url)
            if r.ok:
                break
        except requests.exceptions.ConnectionError:
            pass


@contextlib.contextmanager
def initialize_s3() -> Generator[Dict[str, str], None, None]:
    pytest.importorskip("boto3")
    if "AWS_SECRET_ACCESS_KEY" not in os.environ:
        os.environ["AWS_SECRET_ACCESS_KEY"] = "foo"
    if "AWS_ACCESS_KEY_ID" not in os.environ:
        os.environ["AWS_ACCESS_KEY_ID"] = "foo"
    port = 5555
    proc = subprocess.Popen(
        shlex.split(f"moto_server s3 -p {port}"),
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stdin=subprocess.DEVNULL,
    )
    endpoint_url = f"http://127.0.0.1:{port}/"
    wait_s3_up(endpoint_url)
    try:
        yield {"endpoint_url": endpoint_url}
    finally:
        proc.terminate()
        proc.wait()


@pytest.fixture(autouse=True)
def set_cache(
    tmpdir: pathlib.Path,
    postgresql: psycopg.Connection[Any],
    request: pytest.FixtureRequest,
) -> Generator[str, None, None]:
    if not hasattr(request, "param") or request.param == "file":
        with config.set(
            cache_db_urlpath="sqlite:///" + str(tmpdir / "cacholote.db"),
            cache_files_urlpath=str(tmpdir / "cache_files"),
        ):
            yield "file"
    elif request.param == "cads":
        pytest.importorskip("s3fs")
        botocore_session = pytest.importorskip("botocore.session")

        test_bucket_name = "test-bucket"
        with initialize_s3() as client_kwargs:
            session = botocore_session.Session()
            client = session.create_client("s3", **client_kwargs)
            client.create_bucket(Bucket=test_bucket_name)
            with config.set(
                cache_db_urlpath=(
                    f"postgresql+psycopg2://{postgresql.info.user}:@{postgresql.info.host}:"
                    f"{postgresql.info.port}/{postgresql.info.dbname}"
                ),
                cache_files_urlpath=f"s3://{test_bucket_name}",
                cache_files_storage_options=dict(client_kwargs=client_kwargs),
            ):
                yield request.param
    else:
        raise ValueError



"""Database objects."""

# Copyright 2023, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
from typing import Any, Dict, Optional

import sqlalchemy as sa
import sqlalchemy.orm

from . import utils

_DATETIME_MAX = datetime.datetime(
    datetime.MAXYEAR, 12, 31, tzinfo=datetime.timezone.utc
)

ENGINE: Optional[sa.engine.Engine] = None
SESSIONMAKER: Optional[sa.orm.sessionmaker] = None  # type: ignore[type-arg]

Base = sa.orm.declarative_base()


class CacheEntry(Base):
    __tablename__ = "cache_entries"

    id = sa.Column(sa.Integer(), primary_key=True)
    key = sa.Column(sa.String(32))
    expiration = sa.Column(sa.DateTime, default=_DATETIME_MAX)
    result = sa.Column(sa.JSON)
    timestamp = sa.Column(
        sa.DateTime,
        default=utils.utcnow,
        onupdate=utils.utcnow,
    )
    counter = sa.Column(sa.Integer)
    tag = sa.Column(sa.String)

    @property
    def _result_as_string(self) -> str:
        return json.dumps(self.result)

    def __repr__(self) -> str:
        return f"CacheEntry(id={self.id!r}, key={self.key!r}, expiration={self.expiration!r})"


@sa.event.listens_for(CacheEntry, "before_insert")
def set_expiration_to_max(
    mapper: sa.orm.Mapper[CacheEntry],
    connection: sa.Connection,
    target: CacheEntry,
) -> None:
    target.expiration = target.expiration or _DATETIME_MAX
    if target.expiration < utils.utcnow():
        raise ValueError(f"Expiration date has passed. {target.expiration=}")


def _commit_or_rollback(session: sa.orm.Session) -> None:
    try:
        session.commit()
    finally:
        session.rollback()


def _set_engine_and_session(
    cache_db_urlpath: str, create_engine_kwargs: Dict[str, Any]
) -> None:
    global ENGINE, SESSIONMAKER
    ENGINE = sa.create_engine(cache_db_urlpath, future=True, **create_engine_kwargs)
    Base.metadata.create_all(ENGINE)
    SESSIONMAKER = sa.orm.sessionmaker(ENGINE)



import contextlib
import datetime
import pathlib
import time
from typing import Any, Literal, Optional, Sequence

import fsspec
import pytest
import structlog

from cacholote import cache, clean, config, utils

does_not_raise = contextlib.nullcontext


@cache.cacheable
def open_url(url: pathlib.Path) -> fsspec.spec.AbstractBufferedFile:
    with fsspec.open(url) as f:
        return f


@pytest.mark.parametrize("method", ["LRU", "LFU"])
@pytest.mark.parametrize("set_cache", ["file", "cads"], indirect=True)
def test_clean_cache_files(
    tmpdir: pathlib.Path,
    set_cache: str,
    method: Literal["LRU", "LFU"],
) -> None:
    con = config.get().engine.raw_connection()
    cur = con.cursor()
    fs, dirname = utils.get_cache_files_fs_dirname()

    # Create files
    for algorithm in ("LRU", "LFU"):
        filename = tmpdir / f"{algorithm}.txt"
        fsspec.filesystem("file").pipe_file(filename, b"1")

    # Copy to cache
    (lru_path,) = {open_url(tmpdir / "LRU.txt").path for _ in range(2)}
    lfu_path = open_url(tmpdir / "LFU.txt").path
    assert set(fs.ls(dirname)) == {lru_path, lfu_path}

    # Do not clean
    clean.clean_cache_files(2, method=method)
    cur.execute("SELECT COUNT(*) FROM cache_entries", ())
    assert cur.fetchone() == (fs.du(dirname),) == (2,)

    # Delete one file
    clean.clean_cache_files(1, method=method)
    cur.execute("SELECT COUNT(*) FROM cache_entries", ())
    assert cur.fetchone() == (fs.du(dirname),) == (1,)
    assert not fs.exists(lru_path if method == "LRU" else lfu_path)


@pytest.mark.parametrize("delete_unknown_files", [True, False])
def test_delete_unknown_files(tmpdir: pathlib.Path, delete_unknown_files: bool) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()

    # Create file
    tmpfile = tmpdir / "test.txt"
    fsspec.filesystem("file").pipe_file(tmpfile, b"1")

    # Copy to cache
    cached_file = open_url(tmpfile).path

    # Add unknown
    fs.put(str(tmpfile), f"{dirname}/unknown.txt")

    # Clean one file
    clean.clean_cache_files(1, delete_unknown_files=delete_unknown_files)
    if delete_unknown_files:
        assert fs.ls(dirname) == [cached_file]
    else:
        assert fs.ls(dirname) == [f"{dirname}/unknown.txt"]


@pytest.mark.parametrize(
    "recursive,raises,final_size",
    [
        (True, does_not_raise(), 0),
        (False, pytest.raises((PermissionError, IsADirectoryError, ValueError)), 1),
    ],
)
def test_delete_unknown_dirs(
    recursive: bool, raises: contextlib.nullcontext, final_size: int  # type: ignore[type-arg]
) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()
    fs.mkdir(f"{dirname}/unknown")
    fs.touch(f"{dirname}/unknown/unknown.txt")
    with raises:
        clean.clean_cache_files(0, delete_unknown_files=True, recursive=recursive)
    assert len(fs.ls(dirname)) == final_size


@pytest.mark.parametrize("lock_validity_period", [None, 0])
def test_clean_locked_files(
    tmpdir: pathlib.Path, lock_validity_period: Optional[float]
) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()

    # Create file
    tmpfile = tmpdir / "test.txt"
    fsspec.filesystem("file").pipe_file(tmpfile, b"1")

    # Copy to cache
    cached_file = open_url(tmpfile).path

    # Add unknown and lock
    fs.put(str(tmpfile), f"{dirname}/unknown.txt")
    fs.touch(f"{dirname}/unknown.txt.lock")

    # Clean one file
    clean.clean_cache_files(
        1, delete_unknown_files=True, lock_validity_period=lock_validity_period
    )
    if lock_validity_period == 0:
        assert fs.ls(dirname) == [cached_file]
    else:
        assert set(fs.ls(dirname)) == {
            f"{dirname}/unknown.txt",
            f"{dirname}/unknown.txt.lock",
        }


@pytest.mark.parametrize(
    "tags_to_clean, tags_to_keep, cleaned",
    [
        ({None, "1"}, None, {None, "1"}),
        ({None, "2"}, None, {None, "2"}),
        ({"1", "2"}, None, {"1", "2"}),
        (None, {None}, {"1", "2"}),
        (None, {"1"}, {None, "2"}),
        (None, {"2"}, {None, "1"}),
    ],
)
def test_clean_tagged_files(
    tmpdir: pathlib.Path,
    tags_to_clean: Optional[Sequence[Optional[str]]],
    tags_to_keep: Optional[Sequence[Optional[str]]],
    cleaned: Sequence[Optional[str]],
) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()

    expected_ls = []
    for tag in [None, "1", "2"]:
        tmpfile = tmpdir / f"test_{tag}.txt"
        fsspec.filesystem("file").pipe_file(tmpfile, b"1")
        with config.set(tag=tag):
            cached_file = open_url(tmpfile).path
        if tag not in cleaned:
            expected_ls.append(cached_file)

    clean.clean_cache_files(1, tags_to_clean=tags_to_clean, tags_to_keep=tags_to_keep)
    assert fs.ls(dirname) == expected_ls


def test_clean_tagged_files_wrong_args() -> None:
    with pytest.raises(
        ValueError,
        match="tags_to_clean/keep are mutually exclusive.",
    ):
        clean.clean_cache_files(1, tags_to_keep=[], tags_to_clean=[])


@pytest.mark.parametrize("wrong_type", ["1", [1]])
def test_clean_tagged_files_wrong_types(wrong_type: Any) -> None:
    raises = pytest.raises(
        TypeError,
        match="tags_to_clean/keep must be None or a sequence of str/None.",
    )
    with raises:
        clean.clean_cache_files(1, tags_to_keep=wrong_type)
    with raises:
        clean.clean_cache_files(1, tags_to_clean=wrong_type)


def test_delete_cache_entry_and_files(tmpdir: pathlib.Path) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()

    # Create file
    tmpfile = tmpdir / "test.txt"
    fsspec.filesystem("file").pipe_file(tmpfile, b"old")

    # Copy to cache
    open_url(tmpfile)

    # Change tmp file
    fsspec.filesystem("file").pipe_file(tmpfile, b"new")
    assert open_url(tmpfile).read() == b"old"

    # Delete cache entry
    clean.delete(open_url, tmpfile)
    assert fs.ls(dirname) == []

    # Cache again
    assert open_url(tmpfile).read() == b"new"
    assert len(fs.ls(dirname)) == 1


@pytest.mark.parametrize("check_expiration", [True, False])
@pytest.mark.parametrize("try_decode", [True, False])
def test_clean_invalid_cache_entries(
    tmpdir: pathlib.Path, check_expiration: bool, try_decode: bool
) -> None:
    fs, dirname = utils.get_cache_files_fs_dirname()

    # Valid cache file
    fsspec.filesystem("file").pipe_file(tmpdir / "valid.txt", b"1")
    valid = open_url(tmpdir / "valid.txt").path

    # Corrupted cache file
    fsspec.filesystem("file").pipe_file(tmpdir / "corrupted.txt", b"1")
    corrupted = open_url(tmpdir / "corrupted.txt").path
    fs.touch(corrupted)

    # Expired cache file
    fsspec.filesystem("file").pipe_file(tmpdir / "expired.txt", b"1")
    with config.set(expiration=utils.utcnow() + datetime.timedelta(seconds=0.2)):
        expired = open_url(tmpdir / "expired.txt").path
    time.sleep(0.2)

    # Clean
    clean.clean_invalid_cache_entries(
        check_expiration=check_expiration, try_decode=try_decode
    )

    # Check files
    fs, dirname = utils.get_cache_files_fs_dirname()
    assert valid in fs.ls(dirname)
    assert (
        corrupted not in fs.ls(dirname) if try_decode else corrupted in fs.ls(dirname)
    )
    assert (
        expired not in fs.ls(dirname) if check_expiration else expired in fs.ls(dirname)
    )

    # Check database
    con = config.get().engine.raw_connection()
    cur = con.cursor()
    cur.execute("SELECT COUNT(*) FROM cache_entries", ())
    assert cur.fetchone() == (3 - check_expiration - try_decode,)


def test_cleaner_logging(
    capsys: pytest.CaptureFixture[str], tmpdir: pathlib.Path
) -> None:
    # Cache file and create unknown
    tmpfile = tmpdir / "test.txt"
    fsspec.filesystem("file").pipe_file(tmpfile, b"1")
    cached_file = open_url(tmpfile)
    fs, dirname = utils.get_cache_files_fs_dirname()
    fs.pipe_file(f"{dirname}/unknown.txt", b"1")

    # Clean
    config.set(logger=structlog.get_logger())
    clean.clean_cache_files(0, delete_unknown_files=True)
    captured = iter(capsys.readouterr().out.splitlines())

    line = next(captured)
    assert "get disk usage of cache files" in line

    line = next(captured)
    assert "get unknown files" in line

    line = next(captured)
    assert "delete unknown" in line
    assert "recursive=False" in line
    assert f"urlpath=file://{dirname}/unknown.txt" in line

    line = next(captured)
    assert "check cache files total size" in line
    assert "size=1" in line

    line = next(captured)
    assert "delete cache entry" in line
    assert "cache_entry=" in line

    line = next(captured)
    assert "delete cache file" in line
    assert f"urlpath=file://{cached_file.path}" in line

    line = next(captured)
    assert "check cache files total size" in line
    assert "size=0" in line



"""Decode JSON data."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
import json
from typing import Any, Callable, Dict, List, Union


def import_object(fully_qualified_name: str) -> Any:
    """Import python objects defined by fully qualified names (``'module:qualname'``)."""
    # FIXME: apply exclude/include-rules to ``fully_qualified_name``
    if ":" not in fully_qualified_name:
        raise ValueError(f"{fully_qualified_name!r} not in the form 'module:qualname'")
    module_name, _, object_name = fully_qualified_name.partition(":")
    obj = importlib.import_module(module_name)
    for attr_name in object_name.split("."):
        obj = getattr(obj, attr_name)
    return obj


class DecodeError(Exception):
    pass


def decode_python_object(obj: Dict[str, Any]) -> Any:
    if obj.get("type") == "python_object" and "fully_qualified_name" in obj:
        return import_object(obj["fully_qualified_name"])
    return None


def decode_python_call(obj: Dict[str, Any]) -> Any:
    if obj.get("type") == "python_call" and "callable" in obj:
        if callable(obj["callable"]):
            func = obj["callable"]
        else:
            func = import_object(obj["callable"])
        args = obj.get("args", ())
        kwargs = obj.get("kwargs", {})
        return func(*args, **kwargs)
    return None


FILECACHE_DECODERS: List[Callable[[Dict[str, Any]], Any]] = [
    decode_python_object,
    decode_python_call,
]


def object_hook(obj: Dict[str, Any]) -> Any:
    """Decode deserialized JSON data (``dict``)."""
    for decoder in reversed(FILECACHE_DECODERS):
        try:
            result = decoder(obj)
            if result is not None:
                return result
        except Exception as ex:
            raise DecodeError(ex) from ex

    return obj


def loads(obj: Union[str, bytes, bytearray], **kwargs: Any) -> Any:
    """Decode serialized JSON data to a python object.

    Parameters
    ----------
    obj: str, bytes, bytearray
        Serialized JSON data.
    **kwargs: Any
        Keyword arguments for ``json.loads``

    Returns
    -------
    Any
    """
    return json.loads(obj, object_hook=object_hook, **kwargs)



"""Functions to clean cache database and files."""

# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import datetime
import functools
import json
import posixpath
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Set, Union

import fsspec
import packaging.version
import sqlalchemy as sa
import sqlalchemy.orm

from . import config, database, decode, encode, extra_encoders, utils


def _delete_cache_file(
    obj: Dict[str, Any],
    session: Optional[sa.orm.Session] = None,
    cache_entry: Optional[database.CacheEntry] = None,
    sizes: Optional[Dict[str, int]] = None,
    dry_run: bool = False,
) -> Any:
    logger = config.get().logger

    if {"type", "callable", "args", "kwargs"} == set(obj) and obj["callable"] in (
        "cacholote.extra_encoders:decode_xr_dataset",
        "cacholote.extra_encoders:decode_io_object",
    ):
        sizes = sizes or {}
        cache_fs, cache_dirname = utils.get_cache_files_fs_dirname()
        cache_dirname = cache_fs.unstrip_protocol(cache_dirname)

        fs, urlpath = extra_encoders._get_fs_and_urlpath(*obj["args"][:2])
        urlpath = fs.unstrip_protocol(urlpath)

        if posixpath.dirname(urlpath) == cache_dirname:
            sizes.pop(urlpath, None)
            if not dry_run:
                if session and cache_entry:
                    logger.info("delete cache entry", cache_entry=cache_entry)
                    session.delete(cache_entry)
                    database._commit_or_rollback(session)

                if fs.exists(urlpath):
                    logger.info("delete cache file", urlpath=urlpath)
                    fs.rm(
                        urlpath,
                        recursive=obj["args"][0]["type"] == "application/vnd+zarr",
                    )

    return obj


def _delete_cache_entry(
    session: sa.orm.Session, cache_entry: database.CacheEntry
) -> None:
    # First, delete database entry
    session.delete(cache_entry)
    database._commit_or_rollback(session)
    # Then, delete files
    json.loads(cache_entry._result_as_string, object_hook=_delete_cache_file)


def delete(
    func_to_del: Union[str, Callable[..., Any]], *args: Any, **kwargs: Any
) -> None:
    """Delete function previously cached.

    Parameters
    ----------
    func_to_del: callable, str
        Function to delete from cache
    *args: Any
        Argument of functions to delete from cache
    **kwargs: Any
        Keyword arguments of functions to delete from cache
    """
    hexdigest = encode._hexdigestify_python_call(func_to_del, *args, **kwargs)
    with config.get().sessionmaker() as session:
        for cache_entry in session.scalars(
            sa.select(database.CacheEntry).filter(database.CacheEntry.key == hexdigest)
        ):
            _delete_cache_entry(session, cache_entry)


class _Cleaner:
    def __init__(self) -> None:
        self.logger = config.get().logger
        self.fs, self.dirname = utils.get_cache_files_fs_dirname()

        urldir = self.fs.unstrip_protocol(self.dirname)

        self.logger.info("get disk usage of cache files")
        self.sizes: Dict[str, int] = collections.defaultdict(lambda: 0)
        for path, size in self.fs.du(self.dirname, total=False).items():
            # Group dirs
            urlpath = self.fs.unstrip_protocol(path)
            basename, *_ = urlpath.replace(urldir, "", 1).strip("/").split("/")
            if basename:
                self.sizes[posixpath.join(urldir, basename)] += size

    @property
    def size(self) -> int:
        sum_sizes = sum(self.sizes.values())
        self.logger.info("check cache files total size", size=sum_sizes)
        return sum_sizes

    def stop_cleaning(self, maxsize: int) -> bool:
        return self.size <= maxsize

    def get_unknown_files(self, lock_validity_period: Optional[float]) -> Set[str]:
        self.logger.info("get unknown files")

        utcnow = (
            utils.utcnow()
            if packaging.version.parse(fsspec.__version__)
            >= packaging.version.parse("2023.9.0")
            else datetime.datetime.utcnow()
        )
        files_to_skip = []
        for urlpath in self.sizes:
            if urlpath.endswith(".lock"):
                delta = utcnow - self.fs.modified(urlpath)
                if lock_validity_period is None or delta < datetime.timedelta(
                    seconds=lock_validity_period
                ):
                    files_to_skip.append(urlpath)
                    files_to_skip.append(urlpath.rsplit(".lock", 1)[0])

        unknown_sizes = {k: v for k, v in self.sizes.items() if k not in files_to_skip}
        if unknown_sizes:
            with config.get().sessionmaker() as session:
                for cache_entry in session.scalars(sa.select(database.CacheEntry)):
                    json.loads(
                        cache_entry._result_as_string,
                        object_hook=functools.partial(
                            _delete_cache_file,
                            sizes=unknown_sizes,
                            dry_run=True,
                        ),
                    )
        return set(unknown_sizes)

    def delete_unknown_files(
        self, lock_validity_period: Optional[float], recursive: bool
    ) -> None:
        for urlpath in self.get_unknown_files(lock_validity_period):
            self.sizes.pop(urlpath)
            if self.fs.exists(urlpath):
                self.logger.info("delete unknown", urlpath=urlpath, recursive=recursive)
                self.fs.rm(urlpath, recursive=recursive)

    @staticmethod
    def check_tags(*args: Any) -> None:
        if None not in args:
            raise ValueError("tags_to_clean/keep are mutually exclusive.")
        for tags in args:
            if tags is not None and (
                not isinstance(tags, (list, set, tuple))
                or not all(tag is None or isinstance(tag, str) for tag in tags)
            ):
                raise TypeError(
                    "tags_to_clean/keep must be None or a sequence of str/None."
                )

    def delete_cache_files(
        self,
        maxsize: int,
        method: Literal["LRU", "LFU"],
        tags_to_clean: Optional[Sequence[Optional[str]]],
        tags_to_keep: Optional[Sequence[Optional[str]]],
    ) -> None:
        self.check_tags(tags_to_clean, tags_to_keep)

        # Filters
        filters = []
        if tags_to_keep is not None:
            filters.append(
                sa.or_(
                    database.CacheEntry.tag.not_in(tags_to_keep),
                    database.CacheEntry.tag.is_not(None)
                    if None in tags_to_keep
                    else database.CacheEntry.tag.is_(None),
                )
            )
        elif tags_to_clean is not None:
            filters.append(
                sa.or_(
                    database.CacheEntry.tag.in_(tags_to_clean),
                    database.CacheEntry.tag.is_(None)
                    if None in tags_to_clean
                    else database.CacheEntry.tag.is_not(None),
                )
            )

        # Sorters
        sorters: List[sa.orm.InstrumentedAttribute[Any]] = []
        if method == "LRU":
            sorters.extend([database.CacheEntry.timestamp, database.CacheEntry.counter])
        elif method == "LFU":
            sorters.extend([database.CacheEntry.counter, database.CacheEntry.timestamp])
        else:
            raise ValueError(f"{method=} is invalid. Choose either 'LRU' or 'LFU'.")
        sorters.append(database.CacheEntry.expiration)

        # Clean database files
        if self.stop_cleaning(maxsize):
            return
        with config.get().sessionmaker() as session:
            for cache_entry in session.scalars(
                sa.select(database.CacheEntry).filter(*filters).order_by(*sorters)
            ):
                json.loads(
                    cache_entry._result_as_string,
                    object_hook=functools.partial(
                        _delete_cache_file,
                        session=session,
                        cache_entry=cache_entry,
                        sizes=self.sizes,
                    ),
                )
                if self.stop_cleaning(maxsize):
                    return

        raise ValueError(
            f"Unable to clean {self.dirname!r}. Final size: {self.size!r}. Expected size: {maxsize!r}"
        )


def clean_cache_files(
    maxsize: int,
    method: Literal["LRU", "LFU"] = "LRU",
    delete_unknown_files: bool = False,
    recursive: bool = False,
    lock_validity_period: Optional[float] = None,
    tags_to_clean: Optional[Sequence[Optional[str]]] = None,
    tags_to_keep: Optional[Sequence[Optional[str]]] = None,
) -> None:
    """Clean cache files.

    Parameters
    ----------
    maxsize: int
        Maximum total size of cache files (bytes).
    method: str, default: "LRU"
        * LRU: Last Recently Used
        * LFU: Least Frequently Used
    delete_unknown_files: bool, default: False
        Delete all files that are not registered in the cache database.
    recursive: bool, default: False
        Whether to delete unknown directories or not
    lock_validity_period: float, optional, default: None
        Validity period of lock files in seconds. Expired locks will be deleted.
    tags_to_clean, tags_to_keep: sequence of strings/None, optional, default: None
        Tags to clean/keep. If None, delete all cache entries.
        To delete/keep untagged entries, add None in the sequence (e.g., [None, 'tag1', ...]).
        tags_to_clean and tags_to_keep are mutually exclusive.
    """
    cleaner = _Cleaner()

    if delete_unknown_files:
        cleaner.delete_unknown_files(lock_validity_period, recursive)

    cleaner.delete_cache_files(
        maxsize=maxsize,
        method=method,
        tags_to_clean=tags_to_clean,
        tags_to_keep=tags_to_keep,
    )


def clean_invalid_cache_entries(
    check_expiration: bool = True, try_decode: bool = False
) -> None:
    """Clean invalid cache entries.

    Parameters
    ----------
    check_expiration: bool
        Whether or not to delete expired entries
    try_decode: bool
        Whether or not to delete entries that raise DecodeError (this can be slow!)
    """
    filters = []
    if check_expiration:
        filters.append(database.CacheEntry.expiration <= utils.utcnow())
    if filters:
        with config.get().sessionmaker() as session:
            for cache_entry in session.scalars(
                sa.select(database.CacheEntry).filter(*filters)
            ):
                _delete_cache_entry(session, cache_entry)

    if try_decode:
        with config.get().sessionmaker() as session:
            for cache_entry in session.scalars(sa.select(database.CacheEntry)):
                try:
                    decode.loads(cache_entry._result_as_string)
                except decode.DecodeError:
                    _delete_cache_entry(session, cache_entry)



"""Global settings."""

# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime
import logging
import pathlib
import tempfile
from types import TracebackType
from typing import Any, Dict, Literal, Optional, Tuple, Type, Union

import fsspec
import pydantic
import pydantic_settings
import sqlalchemy as sa
import sqlalchemy.orm
import structlog

from . import database

_SETTINGS: Optional[Settings] = None
_DEFAULT_CACHE_DIR = pathlib.Path(tempfile.gettempdir()) / "cacholote"
_DEFAULT_CACHE_DIR.mkdir(exist_ok=True)
_DEFAULT_LOGGER = structlog.get_logger(
    wrapper_class=structlog.make_filtering_bound_logger(logging.WARNING)
)

_CONFIG_NOT_SET_MSG = (
    "Configuration settings have not been set. Run `cacholote.config.reset()`."
)


class Settings(pydantic_settings.BaseSettings):
    use_cache: bool = True
    cache_db_urlpath: str = f"sqlite:///{_DEFAULT_CACHE_DIR / 'cacholote.db'}"
    create_engine_kwargs: Dict[str, Any] = {}
    cache_files_urlpath: str = f"{_DEFAULT_CACHE_DIR / 'cache_files'}"
    cache_files_urlpath_readonly: Optional[str] = None
    cache_files_storage_options: Dict[str, Any] = {}
    xarray_cache_type: Literal[
        "application/netcdf", "application/x-grib", "application/vnd+zarr"
    ] = "application/netcdf"
    io_delete_original: bool = False
    raise_all_encoding_errors: bool = False
    expiration: Optional[datetime.datetime] = None
    tag: Optional[str] = None
    return_cache_entry: bool = False
    logger: Union[
        structlog.BoundLogger, structlog._config.BoundLoggerLazyProxy
    ] = _DEFAULT_LOGGER
    lock_timeout: Optional[float] = None

    @pydantic.field_validator("create_engine_kwargs")
    def validate_create_engine_kwargs(
        cls: pydantic_settings.BaseSettings, create_engine_kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        poolclass = create_engine_kwargs.get("poolclass")
        if isinstance(poolclass, str):
            create_engine_kwargs["poolclass"] = getattr(sa.pool, poolclass)
        return create_engine_kwargs

    @pydantic.field_validator("return_cache_entry")
    def validate_return_cache_entry(
        cls: pydantic_settings.BaseSettings,
        return_cache_entry: bool,
        info: pydantic.FieldValidationInfo,
    ) -> bool:
        if return_cache_entry is True and info.data["use_cache"] is False:
            raise ValueError(
                "`use_cache` must be True when `return_cache_entry` is True"
            )
        return return_cache_entry

    @pydantic.field_validator("expiration")
    def validate_expiration(
        cls: pydantic_settings.BaseSettings, expiration: Optional[datetime.datetime]
    ) -> Optional[datetime.datetime]:
        if expiration is not None and expiration.tzinfo is None:
            raise ValueError(f"Expiration is missing the timezone info. {expiration=}")
        return expiration

    def make_cache_dir(self) -> None:
        fs, _, (urlpath, *_) = fsspec.get_fs_token_paths(
            self.cache_files_urlpath,
            storage_options=self.cache_files_storage_options,
        )
        fs.mkdirs(urlpath, exist_ok=True)

    def set_engine_and_session(self, force_reset: bool = False) -> None:
        if (
            force_reset
            or database.ENGINE is None
            or database.SESSIONMAKER is None
            or str(database.ENGINE.url) != self.cache_db_urlpath
        ):
            database._set_engine_and_session(
                self.cache_db_urlpath, self.create_engine_kwargs
            )

    @property
    def engine(self) -> sa.engine.Engine:
        if database.ENGINE is None:
            raise ValueError(_CONFIG_NOT_SET_MSG)
        return database.ENGINE

    @property
    def sessionmaker(self) -> sa.orm.sessionmaker:  # type: ignore[type-arg]
        if database.SESSIONMAKER is None:
            raise ValueError(_CONFIG_NOT_SET_MSG)
        return database.SESSIONMAKER

    model_config = pydantic_settings.SettingsConfigDict(
        case_sensitive=False, env_prefix="cacholote_"
    )


class set:
    """Customize cacholote settings.

    It is possible to use it either as a context manager, or to configure global settings.

    Parameters
    ----------
    use_cache: bool, default: True
        Enable/disable cache.
    cache_db_urlpath: str, default:"sqlite:////system_tmp_dir/cacholote/cacholote.db"
        URL for cache database (driver://user:pass@host/database).
    create_engine_kwargs: dict, default: {}
        Keyword arguments for ``sqlalchemy.create_engine``
    cache_files_urlpath: str, default:"/system_tmp_dir/cacholote/cache_files"
        URL for cache files (protocol://location).
    cache_files_storage_options: dict, default: {}
        ``fsspec`` storage options for storing cache files.
    cache_files_urlpath_readonly: str, None, default: None
        URL for cache files accessible in read-only mode.
        None: same as ``cache_files_urlpath``
    xarray_cache_type: {"application/netcdf", "application/x-grib", "application/vnd+zarr"}, \
        default: "application/netcdf"
        Type for ``xarray`` cache files.
    io_delete_original: bool, default: False
        Whether to delete the original copy of cached files.
    raise_all_encoding_errors: bool, default: False
        Raise an error if an encoder does not work (i.e., do not return results).
    expiration: datetime, optional, default: None
        Expiration for cached results.
    tag: str, optional, default: None
        Tag for the cache entry. If None, do NOT tag.
        Note that existing tags are overwritten.
    return_cache_entry: bool, default: False
        Whether to return the cache database entry rather than decoded results.
    lock_timeout: fload, optional, default: None
        Time to wait before raising an error if a cache file is locked.
    """

    def __init__(self, **kwargs: Any):
        self._old_engine = database.ENGINE
        self._old_session = database.SESSIONMAKER
        self._old_settings = get()

        global _SETTINGS
        _SETTINGS = Settings(**{**self._old_settings.model_dump(), **kwargs})
        _SETTINGS.make_cache_dir()
        _SETTINGS.set_engine_and_session(
            self._old_settings.create_engine_kwargs != _SETTINGS.create_engine_kwargs
        )

    def __enter__(self) -> Settings:
        return get()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        database.ENGINE = self._old_engine
        database.SESSIONMAKER = self._old_session

        global _SETTINGS
        _SETTINGS = self._old_settings


def reset(env_file: Optional[Union[str, Tuple[str]]] = None) -> None:
    """Reset cacholote settings.

    Priority:
    1. Environment variables with prefix `CACHOLOTE_`
    2. Dotenv file(s)
    3. Cacholote defaults

    Parameters
    ----------
    env_file: str, tuple[str], default=None
        Dot env file(s).
    """
    global _SETTINGS
    _SETTINGS = Settings(_env_file=env_file)  # type: ignore[call-arg]
    set()


def get() -> Settings:
    """Get cacholote settings."""
    if _SETTINGS is None:
        raise ValueError(_CONFIG_NOT_SET_MSG)
    return _SETTINGS.model_copy()


reset()



"""Additional encoders that need optional dependencies."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import functools
import hashlib
import inspect
import io
import mimetypes
import pathlib
import posixpath
import tempfile
import time
from typing import Any, Callable, Dict, Generator, Optional, Tuple, TypeVar, Union, cast

import fsspec
import fsspec.implementations.local
import pydantic

from . import config, encode, utils

try:
    import dask
    import xarray as xr

    _HAS_XARRAY_AND_DASK = True
except ImportError:
    _HAS_XARRAY_AND_DASK = False

try:
    import magic

    _HAS_MAGIC = True
except ImportError:
    _HAS_MAGIC = False

F = TypeVar("F", bound=Callable[..., Any])

_UNION_IO_TYPES = Union[
    io.RawIOBase,
    io.BufferedIOBase,
    io.TextIOBase,
    fsspec.spec.AbstractBufferedFile,
    fsspec.implementations.local.LocalFileOpener,
]


def _add_ext_to_mimetypes() -> None:
    """Add netcdf, grib, and zarr to mimetypes."""
    mimetypes.add_type("application/netcdf", ".nc", strict=True)
    for ext in (".grib", ".grb", ".grb1", ".grb2"):
        mimetypes.add_type("application/x-grib", ext, strict=False)
    mimetypes.add_type("application/vnd+zarr", ".zarr", strict=False)


_add_ext_to_mimetypes()


def _guess_type(
    fs: fsspec.AbstractFileSystem,
    local_path: str,
    default: str = "application/octet-stream",
) -> str:
    content_type: str = fs.info(local_path).get("ContentType", "")
    if content_type:
        return content_type

    filetype, *_ = mimetypes.guess_type(local_path, strict=False)
    if filetype is None and _HAS_MAGIC:
        with fs.open(local_path, "rb") as f:
            filetype = magic.from_buffer(f.read(), mime=True)
    return filetype or default


def _requires_xarray_and_dask(func: F) -> F:
    """Raise an error if `xarray` or `dask` are not installed."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not _HAS_XARRAY_AND_DASK:
            raise ValueError("please install 'xarray' and 'dask'")
        return func(*args, **kwargs)

    return cast(F, wrapper)


@contextlib.contextmanager
def _logging_timer(event: str, **kwargs: Any) -> Generator[float, None, None]:
    logger = config.get().logger
    logger.info(f"start {event}", **kwargs)
    tic = time.perf_counter()
    yield tic
    toc = time.perf_counter()
    kwargs["_".join(event.split() + ["time"])] = toc - tic  # elapsed time
    logger.info(f"end {event}", **kwargs)


class FileInfoModel(pydantic.BaseModel):
    type: str
    href: str
    file_checksum: str = pydantic.Field(..., alias="file:checksum")
    file_size: int = pydantic.Field(..., alias="file:size")
    file_local_path: str = pydantic.Field(..., alias="file:local_path")


def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> Dict[str, Any]:
    href = posixpath.join(
        config.get().cache_files_urlpath_readonly or config.get().cache_files_urlpath,
        posixpath.basename(local_path),
    )
    file_dict = {
        "type": _guess_type(fs, local_path),
        "href": href,
        "file:checksum": f"{fs.checksum(local_path):x}",
        "file:size": fs.size(local_path),
        "file:local_path": local_path,
    }

    return FileInfoModel(**file_dict).model_dump(by_alias=True)


def _get_fs_and_urlpath(
    file_json: Dict[str, Any],
    storage_options: Optional[Dict[str, Any]] = None,
    validate: bool = False,
) -> Tuple[fsspec.AbstractFileSystem, str]:
    urlpath = file_json["file:local_path"]
    if storage_options is None:
        storage_options = config.get().cache_files_storage_options

    if not validate:
        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
        return (fs, urlpath)

    # Attempt to read from local_path
    try:
        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
    except:  # noqa: E722
        pass
    else:
        if fs.exists(urlpath):
            expected = file_json["file:checksum"]
            actual = (
                fs.checksum(urlpath)  # Just for backward compatibility.
                if isinstance(expected, int)
                else f"{fs.checksum(urlpath):x}"
            )
            if expected != actual:
                raise ValueError(f"checksum mismatch: {urlpath=} {expected=} {actual=}")
            config.get().logger.info(
                "retrieve cache file", urlpath=fs.unstrip_protocol(urlpath)
            )
            return (fs, urlpath)

    # Attempt to read from href
    urlpath = file_json["href"]
    fs, *_ = fsspec.get_fs_token_paths(urlpath)
    if fs.exists(urlpath):
        return (fs, urlpath)

    # Nothing worked
    raise ValueError(
        f"No such file or directory: {file_json['file:local_path']!r} nor {file_json['href']!r}"
    )


@_requires_xarray_and_dask
def decode_xr_dataset(
    file_json: Dict[str, Any], storage_options: Dict[str, Any], **kwargs: Any
) -> "xr.Dataset":
    fs, urlpath = _get_fs_and_urlpath(
        file_json, storage_options=storage_options, validate=True
    )

    if file_json["type"] == "application/vnd+zarr":
        filename_or_obj = fs.get_mapper(urlpath)
    else:
        if fs.protocol == "file":
            filename_or_obj = urlpath
        else:
            # Download local copy
            protocols = [fs.protocol] if isinstance(fs.protocol, str) else fs.protocol
            with fsspec.open(
                f"filecache::{urlpath}",
                filecache={"same_names": True},
                **{protocol: fs.storage_options for protocol in protocols},
            ) as of:
                filename_or_obj = of.name
    return xr.open_dataset(filename_or_obj, **kwargs)


def decode_io_object(
    file_json: Dict[str, Any], storage_options: Dict[str, Any], **kwargs: Any
) -> _UNION_IO_TYPES:
    fs, urlpath = _get_fs_and_urlpath(
        file_json, storage_options=storage_options, validate=True
    )
    return fs.open(urlpath, **kwargs)


@_requires_xarray_and_dask
def _maybe_store_xr_dataset(
    obj: "xr.Dataset", fs: fsspec.AbstractFileSystem, urlpath: str, filetype: str
) -> None:
    if filetype == "application/vnd+zarr":
        with utils.FileLock(
            fs, urlpath, timeout=config.get().lock_timeout
        ) as file_exists:
            if not file_exists:
                # Write directly on any filesystem
                mapper = fs.get_mapper(urlpath)
                with _logging_timer("upload", urlpath=fs.unstrip_protocol(urlpath)):
                    obj.to_zarr(mapper, consolidated=True)
    elif not fs.exists(urlpath):
        # Need a tmp local copy to write on a different filesystem
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpfilename = str(pathlib.Path(tmpdirname) / pathlib.Path(urlpath).name)

            with _logging_timer("write tmp file", urlpath=tmpfilename):
                if filetype == "application/netcdf":
                    obj.to_netcdf(tmpfilename)
                elif filetype == "application/x-grib":
                    import cfgrib.xarray_to_grib

                    cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
                else:
                    # Should never get here! xarray_cache_type is checked in config.py
                    raise ValueError(f"type {filetype!r} is NOT supported.")

            _maybe_store_file_object(
                fs if fs.protocol == "file" else fsspec.filesystem("file"),
                tmpfilename,
                fs,
                urlpath,
                io_delete_original=True,
            )


@_requires_xarray_and_dask
def dictify_xr_dataset(obj: "xr.Dataset") -> Dict[str, Any]:
    """Encode a ``xr.Dataset`` to JSON deserialized data (``dict``)."""
    with dask.config.set({"tokenize.ensure-deterministic": True}):
        root = dask.base.tokenize(obj)  # type: ignore[no-untyped-call]

    filetype = config.get().xarray_cache_type
    ext = mimetypes.guess_extension(filetype, strict=False)
    urlpath_out = posixpath.join(config.get().cache_files_urlpath, f"{root}{ext}")

    fs_out, *_ = fsspec.get_fs_token_paths(
        urlpath_out,
        storage_options=config.get().cache_files_storage_options,
    )
    _maybe_store_xr_dataset(obj, fs_out, urlpath_out, filetype)

    file_json = _dictify_file(fs_out, urlpath_out)
    kwargs: Dict[str, Any] = {"chunks": {}}
    if filetype == "application/vnd+zarr":
        kwargs.update({"engine": "zarr", "consolidated": True})

    return encode.dictify_python_call(
        decode_xr_dataset,
        file_json,
        storage_options=config.get().cache_files_storage_options,
        **kwargs,
    )


def _maybe_store_file_object(
    fs_in: fsspec.AbstractFileSystem,
    urlpath_in: str,
    fs_out: fsspec.AbstractFileSystem,
    urlpath_out: str,
    io_delete_original: Optional[bool] = None,
) -> None:
    if io_delete_original is None:
        io_delete_original = config.get().io_delete_original
    with utils.FileLock(
        fs_out, urlpath_out, timeout=config.get().lock_timeout
    ) as file_exists:
        if not file_exists:
            kwargs = {}
            content_type = _guess_type(fs_in, urlpath_in)
            if content_type:
                kwargs["ContentType"] = content_type
            with _logging_timer(
                "upload",
                urlpath=fs_out.unstrip_protocol(urlpath_out),
                size=fs_in.size(urlpath_in),
            ):
                if fs_in == fs_out:
                    if io_delete_original:
                        fs_in.mv(urlpath_in, urlpath_out, **kwargs)
                    else:
                        fs_in.cp(urlpath_in, urlpath_out, **kwargs)
                elif fs_in.protocol == "file":
                    fs_out.put(urlpath_in, urlpath_out, **kwargs)
                else:
                    with fs_in.open(urlpath_in, "rb") as f_in:
                        with fs_out.open(urlpath_out, "wb") as f_out:
                            utils.copy_buffered_file(f_in, f_out)
    if io_delete_original and fs_in.exists(urlpath_in):
        with _logging_timer(
            "remove",
            urlpath=fs_in.unstrip_protocol(urlpath_in),
            size=fs_in.size(urlpath_in),
        ):
            fs_in.rm(urlpath_in)


def _maybe_store_io_object(
    f_in: _UNION_IO_TYPES,
    fs_out: fsspec.AbstractFileSystem,
    urlpath_out: str,
) -> None:
    with utils.FileLock(
        fs_out, urlpath_out, timeout=config.get().lock_timeout
    ) as file_exists:
        if not file_exists:
            f_out = fs_out.open(urlpath_out, "wb")
            with _logging_timer("upload", urlpath=fs_out.unstrip_protocol(urlpath_out)):
                utils.copy_buffered_file(f_in, f_out)


def dictify_io_object(obj: _UNION_IO_TYPES) -> Dict[str, Any]:
    """Encode a file object to JSON deserialized data (``dict``)."""
    cache_files_urlpath = config.get().cache_files_urlpath
    fs_out, *_ = fsspec.get_fs_token_paths(
        cache_files_urlpath,
        storage_options=config.get().cache_files_storage_options,
    )

    if hasattr(obj, "path") or hasattr(obj, "name"):
        urlpath_in = obj.path if hasattr(obj, "path") else obj.name  # type: ignore[union-attr]
        fs_in = getattr(obj, "fs", fsspec.filesystem("file"))
        root = f"{fs_in.checksum(urlpath_in):x}"
        ext = pathlib.Path(urlpath_in).suffix
        urlpath_out = posixpath.join(cache_files_urlpath, f"{root}{ext}")
        _maybe_store_file_object(fs_in, urlpath_in, fs_out, urlpath_out)
    else:
        root = hashlib.md5(f"{hash(obj)}".encode()).hexdigest()  # fsspec uses md5
        urlpath_out = posixpath.join(cache_files_urlpath, root)
        _maybe_store_io_object(obj, fs_out, urlpath_out)

    file_json = _dictify_file(fs_out, urlpath_out)
    params = inspect.signature(open).parameters
    kwargs = {k: getattr(obj, k) for k in params.keys() if hasattr(obj, k)}

    return encode.dictify_python_call(
        decode_io_object,
        file_json,
        storage_options=config.get().cache_files_storage_options,
        **kwargs,
    )


def register_all() -> None:
    """Register extra encoders if optional dependencies are installed."""
    for type_ in (
        io.RawIOBase,
        io.BufferedIOBase,
        io.TextIOBase,
        fsspec.spec.AbstractBufferedFile,
        fsspec.implementations.local.LocalFileOpener,
    ):
        encode.FILECACHE_ENCODERS.append((type_, dictify_io_object))
    if _HAS_XARRAY_AND_DASK:
        encode.FILECACHE_ENCODERS.append((xr.Dataset, dictify_xr_dataset))



import datetime
import pathlib
import time
from typing import Any

import pytest

from cacholote import cache, config, database


def func(a: Any, *args: Any, b: Any = None, **kwargs: Any) -> Any:
    if b is None:
        return locals()
    else:

        class LocalClass:
            pass

        return LocalClass()


@cache.cacheable
def cached_now() -> datetime.datetime:
    return datetime.datetime.now()


@cache.cacheable
def cached_error() -> None:
    raise ValueError("test error")


def test_cacheable(tmpdir: pathlib.Path) -> None:
    con = config.get().engine.raw_connection()
    cur = con.cursor()

    cfunc = cache.cacheable(func)

    for counter in range(1, 3):
        before = datetime.datetime.now(tz=datetime.timezone.utc)
        res = cfunc("test")
        after = datetime.datetime.now(tz=datetime.timezone.utc)
        assert res == {"a": "test", "args": [], "b": None, "kwargs": {}}

        cur.execute(
            "SELECT id, key, expiration, result, counter FROM cache_entries", ()
        )
        assert cur.fetchall() == [
            (
                1,
                "a8260ac3cdc1404aa64a6fb71e853049",
                "9999-12-31 00:00:00.000000",
                '{"a": "test", "b": null, "args": [], "kwargs": {}}',
                counter,
            )
        ]

        cur.execute("SELECT timestamp FROM cache_entries", ())
        (timestamp,) = cur.fetchone() or []
        assert before < datetime.datetime.fromisoformat(timestamp + "+00:00") < after


@pytest.mark.parametrize("raise_all_encoding_errors", [True, False])
def test_encode_errors(tmpdir: pathlib.Path, raise_all_encoding_errors: bool) -> None:
    config.set(raise_all_encoding_errors=raise_all_encoding_errors)

    cfunc = cache.cacheable(func)

    class Dummy:
        pass

    inst = Dummy()

    if raise_all_encoding_errors:
        with pytest.raises(AttributeError):
            cfunc(inst)
    else:
        with pytest.warns(UserWarning, match="can NOT encode python call"):
            res = cfunc(inst)
        assert res == {"a": inst, "args": (), "b": None, "kwargs": {}}

    if raise_all_encoding_errors:
        with pytest.raises(AttributeError):
            cfunc("test", b=1)
    else:
        with pytest.warns(UserWarning, match="can NOT encode output"):
            res = cfunc("test", b=1)
        assert res.__class__.__name__ == "LocalClass"

    # cache-db must be empty
    con = config.get().engine.raw_connection()
    cur = con.cursor()
    cur.execute("SELECT COUNT(*) FROM cache_entries", ())
    assert cur.fetchone() == (0,)


def test_same_args_kwargs() -> None:
    ufunc = cache.cacheable(func)

    con = config.get().engine.raw_connection()
    cur = con.cursor()

    ufunc(1)
    cur.execute("SELECT id, key, counter FROM cache_entries", ())
    assert cur.fetchall() == [(1, "54f546036ae7dccdd0155893189154c0", 1)]

    ufunc(a=1)
    cur.execute("SELECT id, key, counter FROM cache_entries", ())
    assert cur.fetchall() == [(1, "54f546036ae7dccdd0155893189154c0", 2)]


@pytest.mark.parametrize("use_cache", [True, False])
def test_use_cache(use_cache: bool) -> None:
    config.set(use_cache=use_cache)

    if use_cache:
        assert cached_now() == cached_now()
    else:
        assert cached_now() < cached_now()


def test_expiration_and_return_cache_entry() -> None:
    config.set(return_cache_entry=True)
    first: database.CacheEntry = cached_now()  # type: ignore[assignment]
    assert first.id == 1
    assert first.key == "c3d9e414d0d32337c3672cb29b1b3cc9"
    assert first.expiration == datetime.datetime(9999, 12, 31)

    dt = datetime.timedelta(seconds=0.1)
    expiration = datetime.datetime.now(tz=datetime.timezone.utc) + dt
    with config.set(expiration=expiration):
        second: database.CacheEntry = cached_now()  # type: ignore[assignment]
        assert second.result != first.result
        assert second.id == 2
        assert second.key == "c3d9e414d0d32337c3672cb29b1b3cc9"
        assert (
            second.expiration is not None
            and second.expiration.isoformat() + "+00:00" == expiration.isoformat()
        )

    time.sleep(0.1)
    third: database.CacheEntry = cached_now()  # type: ignore[assignment]
    assert third.result == first.result
    assert third.id == 1
    assert third.key == "c3d9e414d0d32337c3672cb29b1b3cc9"
    assert third.expiration == datetime.datetime(9999, 12, 31)


def test_tag(tmpdir: pathlib.Path) -> None:
    con = config.get().engine.raw_connection()
    cur = con.cursor()

    cached_now()
    cur.execute("SELECT tag, counter FROM cache_entries", ())
    assert cur.fetchall() == [(None, 1)]

    with config.set(tag="1"):
        cached_now()
    cur.execute("SELECT tag, counter FROM cache_entries", ())
    assert cur.fetchall() == [("1", 2)]

    with config.set(tag="2"):
        # Overwrite
        cached_now()
    cur.execute("SELECT tag, counter FROM cache_entries", ())
    assert cur.fetchall() == [("2", 3)]

    with config.set(tag=None):
        # Do not overwrite if None
        cached_now()
    cur.execute("SELECT tag, counter FROM cache_entries", ())
    assert cur.fetchall() == [("2", 4)]


def test_cached_error() -> None:
    con = config.get().engine.raw_connection()
    cur = con.cursor()

    with pytest.raises(ValueError, match="test error"):
        cached_error()

    cur.execute("SELECT COUNT(*) FROM cache_entries", ())
    assert cur.fetchone() == (0,)



"""Utilities."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import hashlib

import dataclasses
import datetime
import functools
import hashlib
import io
import time
import warnings
from types import TracebackType
from typing import Any, Optional, Tuple, Type

import fsspec

from . import config


def hexdigestify(text: str) -> str:
    """Convert text to its hash made of hexadecimal digits."""
    hash_req = hashlib.sha3_224(text.encode())
    return hash_req.hexdigest()[:32]


def get_cache_files_fs_dirname() -> Tuple[fsspec.AbstractFileSystem, str]:
    """Return the ``fsspec`` filesystem and directory name where cache files are stored."""
    fs, _, (path,) = fsspec.get_fs_token_paths(
        config.get().cache_files_urlpath,
        storage_options=config.get().cache_files_storage_options,
    )
    fs.invalidate_cache()
    return (fs, path)


def copy_buffered_file(
    f_in: Any,
    f_out: fsspec.spec.AbstractBufferedFile,
    buffer_size: Optional[int] = None,
) -> None:
    """Copy file in chunks.

    Parameters
    ----------
    f_in, f_out: fsspec.spec.AbstractBufferedFile
        Source and destination buffered files.
    buffer_size: int, optional, default=None
        Maximum size for chunks in bytes.
        None: use ``io.DEFAULT_BUFFER_SIZE``
    """
    if buffer_size is None:
        buffer_size = io.DEFAULT_BUFFER_SIZE
    while True:
        data = f_in.read(buffer_size)
        if not data:
            break
        f_out.write(data if isinstance(data, bytes) else data.encode())


@dataclasses.dataclass
class FileLock:
    fs: fsspec.AbstractFileSystem  # fsspec file system
    urlpath: str  # file to lock
    timeout: Optional[float]  # lock timeout in seconds

    @functools.cached_property
    def lockfile(self) -> str:
        return self.urlpath + ".lock"

    def acquire(self) -> None:
        self.fs.touch(self.lockfile)

    def release(self) -> None:
        if self.fs.exists(self.lockfile):
            self.fs.rm(self.lockfile)

    @property
    def is_locked(self) -> bool:
        return bool(self.fs.exists(self.lockfile))

    def wait_until_released(self) -> None:
        warned = False
        message = f"{self.urlpath!r} is locked: {self.lockfile!r}"
        start = time.perf_counter()
        while self.is_locked:
            if self.timeout is not None and time.perf_counter() - start > self.timeout:
                raise TimeoutError(message)
            if not warned:
                warnings.warn(message, UserWarning)
                warned = True
            time.sleep(min(1, self.timeout or 1))

    def __enter__(self) -> bool:
        self.wait_until_released()
        self.acquire()
        return bool(self.fs.exists(self.urlpath))

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.release()


def utcnow() -> datetime.datetime:
    """See https://discuss.python.org/t/deprecating-utcnow-and-utcfromtimestamp/26221."""
    return datetime.datetime.now(tz=datetime.timezone.utc)



"""Efficiently cache calls to functions."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import config, database, extra_encoders, utils
from .cache import cacheable
from .clean import clean_cache_files, clean_invalid_cache_entries, delete
from .decode import loads
from .encode import dumps

try:
    # NOTE: the `version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from .version import __version__
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"

extra_encoders.register_all()

__all__ = [
    "__version__",
    "cacheable",
    "clean_cache_files",
    "clean_invalid_cache_entries",
    "config",
    "database",
    "delete",
    "dumps",
    "extra_encoders",
    "loads",
    "utils",
]
